/*
	FT245RL - 8254 Beep Sound Controller
	2016/02/15	Rei_Shinonome
	
	FT245RL Pin Assign

	DB0	:	Serial Data
	DB1	:	Serial Clock
	DB2	:	~WR for 8254
	DB3	:	A0 for 8254
	DB4	:	A1 for 8254
	DB5	:	GATE for 8254
	DB6	:	NC
	DB7	:	NC

	DB0 and DB1 should be connected to a certain SIPO shift register.
	Pararell Out of the shift register should be connected to 8254 D0-7.
	8254 ~RD is not used. It must be Inactive.

*/
#include "ftd2xx.h"

#include <iostream>
#include <fstream>
#include <vector>
#include <list>
#include <string>

#include <stdio.h>
#include <time.h>
#include <Windows.h>
#include <signal.h>
#include <conio.h>

#define BAUDRATE	3000000		//(3000000*4) bps
#define MASK		0xFF		//All Output
#define BITMODE		0x1		//0x1:Asynchronous Bit Bang	0x4:Synchronous BitBang

#define C_BURST		3579545
#define MASTER_CLK	1193182

using namespace std;

struct D_8254 {
	unsigned char low;
	unsigned char high;
} _D_8254[128];

const double BaseFreq[12] = {	//C3-B3
	261.625565,
	277.182631,
	293.664768,
	311.126984,
	329.627556,
	349.228231,
	369.994423,
	391.995436,
	415.304698,
	440.000000,
	466.163762,
	493.883301
};

unsigned char initff[3] = {
	255,
	36,
	4
};

	double m_Freq[128]; //Frequency
	unsigned short int div_C[128]; //Dividing constant

	int tempo = 0;	//BPM
	int parts = 0;	//p[g
	int tick_len = 0;	//ms of 64th note;
	int mute[3] = {0, 0, 0};
	int mutenum[3]={0x32, 0x72, 0xB2};	//Mode1
	int un_mutenum[3]={0x36, 0x76, 0xB6};	//Mode3
	char a[2];	//KeyInWait
	string ofs;	//FileName

	vector<string> array1;	//temporary Array
	vector<vector<int>> tone;	//tone and length
	vector<int> temp;	//buffer for array1 to tone
	string str;	//getline buffer
	ifstream ifs;	//filestream
	long start_t = 0;	//for measuring elapsed time
	long end_t = 0 ;
	unsigned int n = 0;	//for measuring mean BPM

// FOR FT245RL
	DWORD	BytesWritten;
	DWORD	numDevs;
	DWORD	Flags;
	DWORD	ID;
	DWORD	Type;
	DWORD	LocId;

	char	SerialNumber[16];
	char	Description[64];
	unsigned char	wrBits[21];

	FT_HANDLE	ftHandle;
	FT_HANDLE	ftHandleTemp;
	FT_STATUS	ftStatus;
//

	
//vg^Cv錾
	void	SetSignal(int _signame); //Ctrl + Ct
	void	SigHandler(int _signame);	//Ctrl + C
	void	csvToRow();	//CSVsɊUB
	void	csvToArray();	//esƂɁACSVR}؂ŔzɂBie|sj
	void	playCntr();	//w肳ꂽɏ]MIDIm[gio[o͂B
	void	playCntr2();	//2p[g؂ւp
	void	initTable();	//MIDI NOTE NO.8254̑Ή\쐬
	void	initmes();	//FT245RL
	void	Send8254(D_8254 _send, unsigned short int _addr);	//8254֕䑗M Send8254(ׂ, `l)
	int	SetBit(int _org, int _adr);	//8254֑f[^rbg̍쐬 SetBit(ׂ1oCgf[^, `l)
//

void main(int argc, char *argv[]){

	SetSignal(SIGINT);

	cout << "Initializing...";
	initTable();	//8254e[u
	
	system("cls");
	cout << "CSMSP for 8254Board by FT245R" << endl;
	cout << "Rei_Shinonome" << endl;
	cout << "Feb 2016" << endl;
	cout << endl;
	cout << "Opening FT245R..." << endl;

	ftStatus = FT_Open(0, &ftHandle);	//FT245RL

	if (ftStatus == FT_OK){

		printf("Successed!\n");
		initmes();

	}
	else{

		cout << "Failed..." << " Code: " << ftStatus << endl;
		return;

	}

	if (argc > 1){

		ifs.open(argv[1]);

	}
	else{

		cout << "Open File:";
		getline(cin,ofs);
		ifs.open(ofs.c_str());	//CSVǂݍ

	}

	cout << endl;

	if (!ifs){

		if((ofs !="") | (argc >1)){

			cout << "Error!" << endl;

		}

		return;

	}
	else{

		cout << "Opened!" << endl;

	}

	cout << endl;
	
	csvToRow();	//CSVsɊU

	csvToArray();	//esƂɁACSVR}؂ŔzɂBie|sj

	SetBit(0x36, 3);
	ftStatus = FT_Write(ftHandle, wrBits, sizeof(wrBits), &BytesWritten);
	SetBit(0x76, 3);
	ftStatus = FT_Write(ftHandle, wrBits, sizeof(wrBits), &BytesWritten);
	SetBit(0xB6, 3);
	ftStatus = FT_Write(ftHandle, wrBits, sizeof(wrBits), &BytesWritten);//e^C}[h3ɂ

	tempo = atoi(array1[0].c_str());	//BPM擾
	tick_len = (int)(((double)60000 / (tempo * 16)) );	//64ɑ鎞ԁi~bj擾
	cout << "Load File:	";

	if	(argc > 1){

		cout << argv[1];

	}
	else{

		cout << ofs;

	}

	cout << endl;
	cout << "BPM:	" << tempo << endl;
	cout << "64th note time:	" << tick_len << endl;
	cout << "PART:	" << parts << endl;
	cout << endl;
	cout << "READY?" << endl;
	_getch();	//READY?

	timeBeginPeriod(1);	//Windows^C}x1ms

	playCntr(); //t[`̎
	//playCntr2() playCntr()Ɣrgp

	timeEndPeriod(1);	//^C}xɖ߂B
	
	cout << endl;	//Ug̕\
	cout << "Set BPM			:	" << tempo << endl;
	cout << "Quarter Note length	:	" << (n / tone.size()) << "ms" << endl;
	cout << "Measurement BPM		:	" << 60000 / (double)(n / tone.size()) << endl;
	
	ftStatus = FT_Write(ftHandle, initff, sizeof(initff), &BytesWritten);

	FT_Close(ftHandle);	//FT245RLnh
	ifs.close();	//t@CXg[

	cout << endl;
	cout <<  "Closed" << endl;

	return;
}

void SetSignal(int _signame){

	if(signal(_signame, SigHandler) == SIG_ERR){

		cout << "Interreput Setting Error!!!" << endl;
		exit(1);

	}

	return;
}

void SigHandler(int _signame){

	cout << endl;
	cout << "Ctrl + C Interreput Catched" << endl;
	cout << "Releasing Handles..." << endl;
	ftStatus = FT_Write(ftHandle, initff, sizeof(initff), &BytesWritten);
	ftStatus = FT_Close(ftHandle);
	cout << "FT_Close Result:	" << ftStatus << endl;
	ifs.close();
	cout << "Released" << endl;
	exit(0);

}

void initTable(){

	//g̒`
	m_Freq[0] = 0;	//00

	for(int i = 0; i < 12; i++){

		m_Freq[i + 60] = BaseFreq[i]; //C3 - B3`

	}

	int itoffset = 59;

	for(int i = itoffset; i > 0; i--){

		m_Freq[i]=m_Freq[i + 12] / 2; //B2O`

	}

	itoffset = 72;

	for(int i = itoffset; i < 128; i++){

		m_Freq[i]=m_Freq[i - 12] * 2; //C4`

	}

	//܂Ŏg̒`

	//番ւ̕ϊ

	for(int i = 0; i < 14; i++){	//NoteNo.0`14ɂĂ͕͈͊O

		div_C[i] = 65535;

	}

	for(int i = 14; i < 128; i++){

		div_C[i] = (int)((MASTER_CLK / m_Freq[i]) + 0.5);

	}

	//܂ŕւ̕ϊ

	//8254ݗp̊t

	for(int i = 0; i < 128; i++){

		_D_8254[i].low = div_C[i] % 256;
		_D_8254[i].high = div_C[i] / 256;

	}

	//܂8254ݗp̊t
	return;

}

void initmes(){

		FT_CreateDeviceInfoList(&numDevs);
		cout << "Number of devices is " << numDevs << endl;
		FT_GetDeviceInfoDetail(0, &Flags, &Type, &ID, &LocId, SerialNumber, Description, &ftHandleTemp);
		printf("Dev:\n", numDevs);
		printf("	Flags		=0x%x\n",Flags);
		printf("	Type		=0x%x\n",Type);
		printf("	ID		=0x%x\n",ID);
		printf("	LocId		=0x%x\n",LocId);
		printf("	Description	=%s\n",Description);
		printf("	ftHandle	=0x%x\n\n",ftHandleTemp);
		FT_SetBaudRate(ftHandle, BAUDRATE);
		printf("BaudRate:	%d\n",BAUDRATE);
		FT_SetBitMode(ftHandle, MASK, BITMODE);
		printf("DIRECTION:	0x%x\n",MASK);
		printf("MODE:		0x%x\n\n", BITMODE);
		printf("%s is READY.\n\n", Description);
		ftStatus = FT_Write(ftHandle, initff, sizeof(initff), &BytesWritten);
		return;

}

void SetBit(int _org, int _adr){

	int _temp1;
	_temp1 = _org ;
	wrBits[0] = 0x24; //GATE ACTIVE, ADDR 0, WR INACTIVE, SC/SD 0
	wrBits[1] = 0x24;

	for(int i = 0; i < 8  ; i++){

		_temp1 = _temp1 & 0xFF;
		_temp1 = _temp1 / 128;
		wrBits[(i+1)*2] = (_temp1) | ((_adr * 8) | 0x24);	//ADDR SET WR INACTIVE
		wrBits[((i+1)*2)+1] = wrBits[(i+1)*2];
		wrBits[(i+1)*2] = (wrBits[(i+1)*2] & 0x3D);		//GATE ACTIVE ADDR SET WR INACTIVE SC 0
		wrBits[((i+1)*2)+1] = wrBits[((i+1)*2)+1] | 0x02;	//GATE ACTIVE ADDR SET WR INACTIVE SC 1
		_temp1 = _org << (i+1);

	}

	wrBits[18] = (_adr * 8) | 0x24; //ADDR SET, WR INACTIVE, GATE ACIVE
	wrBits[19] = wrBits[18] & 0x3B; //ADDR SET, WR ACTIVE, GATE ACTIVE
	wrBits[20] = wrBits[18];	//WR INACTIVE
	return;

}

void Send8254(D_8254 _send, unsigned short int _addr){

	SetBit(_send.low, _addr);
	ftStatus = FT_Write(ftHandle, wrBits, sizeof(wrBits), &BytesWritten);
	printf("WRITE_LOW:		%d\n",_send.low);
	start_t=clock();
	SetBit(_send.high, _addr);
	ftStatus = FT_Write(ftHandle, wrBits, sizeof(wrBits), &BytesWritten);
	printf("WRITE_HIGH:		%d\n",_send.high);
	return;

}

void csvToRow(){		//CSVsɊUB

	while( getline(ifs, str)){

		array1.push_back( str );

	}//evfɈsB

	return;

}

void csvToArray(){	//esƂɁACSVR}؂ŔzɂBie|sj

	int splitIndex = 0;
	parts = atoi(array1[1].c_str());

	for(unsigned int i=2 ; i < array1.size(); i++){
		
		do{

			splitIndex = array1[i].find_first_of(",");
			temp.push_back(atoi(array1[i].substr(0,splitIndex).c_str()));
			array1[i] = array1[i].substr(splitIndex+1);

		}while( splitIndex != string::npos);//J}ŋ؂tempɓB(Tone1 Tone2 Tone3 Tickj̏

		tone.push_back(temp);//temptoneɂBāATick̒Tone1Tone2Tone3̑gXƂĂ
		temp.clear();

	}
	return;

}

void playCntr(){

	for (unsigned int i=0; i<tone.size(); i++){

		start_t = clock();

		for (int p=0; p<=parts; p++){

			cout << tone[i][p];

			if (p != parts) {

				cout << ",";

			}

		}

		cout << endl;
		cout << (tick_len * (64 / tone[i][parts])) << endl;

		for (int p=0; p<parts; p++){	//~[g

			if ((mute[p] == 1) & (tone[i][p] != 0)){

				SetBit(un_mutenum[p], 3);
				ftStatus = FT_Write(ftHandle, wrBits, sizeof(wrBits), &BytesWritten);
				mute[p] = 0;

			}

		}

		for (int p=0; p<parts; p++){

			if (tone[i][p] !=0){

				Send8254(_D_8254[tone[i][p]], p);

			}
			else{

				SetBit(mutenum[p], 3);
				ftStatus = FT_Write(ftHandle, wrBits, sizeof(wrBits), &BytesWritten);
				mute[p] = 1;

			}

			cout << endl;

		}

		cout << endl;
		//̕ҋ@
		Sleep(tick_len * (64 / tone[i][parts]) );

		end_t = clock();

		//ۂ̏vԕ\
		cout << "REAL:	" << end_t-start_t << endl << endl;

		n = n + (((end_t - start_t) * tone[i][parts]) / 4);
		
	}

	return;

}

void playCntr2(){

	for (unsigned int i=0; i<tone.size(); i++){

		start_t = clock();

		//tm[g̏ڍׂ\
		cout << tone[i][0] << "," << tone[i][1] << "," << tone[i][2] << endl;
		cout << (tick_len * (64 / tone[i][parts])) << endl;
		
		//]
		for(int j = 0; j < (32 / tone[i][parts]); j++){

			Send8254(_D_8254[tone[i][0]], 0);
			cout << endl;
			Sleep(tick_len * (64 / 64) );
			Send8254(_D_8254[tone[i][1]], 0);
			cout << endl;
			//̕ҋ@
			Sleep(tick_len * (64 / 64) );
		
}

		end_t = clock();

		//ۂ̏vԕ\
		cout << "REAL:	" << end_t-start_t << endl << endl;

		n = n + (((end_t - start_t) * tone[i][parts]) / 4 );
		
	}

	return;

}